// 計測する画像に関するクラス
class CurrentImage {
  PImage  img;
  int     max_measure;             // 最大計測点数（この数値の数だけ計測が完了したら自動終了（次のファイルを選択），0の場合は意図的に終了するまで計測を続ける
  String  fullpathname;            // 開いている画像ファイルのフルパス名
  double  scale;                   // 画像の拡大率 ：注意　計測中にも変更できるが，計測精度が落ちる恐れがある．読み込み前に変更するか，再読み込みする
  int     org_width, org_height;   // 画像のオリジナルサイズ
  boolean folder_mode;             // フォルダーモード時，true;
  File[]  files;                   // フォルダーモード：ファイルリスト
  int     f_no;                    // フォルダーモード時の読みだし位置
  int     filter_no;               // -1 : なし，0 : GRAY, 1 : INVERT
  String  folder_mode_folder;      // フォルダーモード時のフォルダー
  
  int     _counter;                // 計測済みデータ数

  // クラス
  PointData []  measured = {};     // 計測データ
  Plot          plotMark;
  
  // コンストラクター
  CurrentImage() {
    img                =  null;
    fullpathname       =  null;
    max_measure        =  Max_measure;
    _counter           =  0;
    scale              =  Initial_scale;
    org_width          = -1;
    org_height         = -1;
    // フォルダーモード関係
    folder_mode        =  false;
    files              =  null;
    f_no               =  0;
    filter_no          = -1;
    folder_mode_folder = "";
    
    // プロットの初期化
    plotMark = new Plot();
  }
  // 画像ファイルの読み込み
  PImage load(String path) {
    boolean reload = false;
    PImage img0;
    if (path == null) reload = true;
    if (! reload) fullpathname = path;
    if ((img0 = loadImage(fullpathname)) == null) return null;  // 読み込み失敗時

    // 読み込み完了
    img = img0;
    shortMessage.set_message("Loaded image: " + fullpathname);
    // フィルターが選択されている場合は先に適用する
    apply_filter();
    // 元画像のサイズを保存
    org_width  = img.width;
    org_height = img.height;
    // 画像を拡大する場合
    if (scale != 1.0) {
      img.resize((int)(img.width * scale), (int)(img.height * scale));
    }
    surface.setSize(img.width, img.height + 40);
    
    if (! reload) _counter  = 0;
    return img;
  }
  // 再読み込み
  PImage reload() {
    if (fullpathname == null) return null;  // 異常な状態
    return load(null);
  }
  
  // 計測データの格納
  int measure(int x, int y) {
    // 画像外（画像の下のショートメッセージ領域）をクリックしてしまった場合
    if (x + y * img.width >= img.pixels.length) return _counter;
    
    boolean skip = false;
    if ((x == -1) && (y == -1)) skip = true;  // 計測スキップ時のフラグ
    
    if ((max_measure > 0) && (_counter >= max_measure)) return _counter;  // 異常な状態
    // 計測ポイント格納領域を追加（back対応）
    if (_counter >= measured.length) {
      PointData pd = new PointData();
      measured = (PointData [])append(measured, pd);
    }
    measured[_counter].px = x;
    measured[_counter].py = y;
    if (! skip) measured[_counter].get_color(img);
    int c = ci.measured[_counter].c;
    if (skip) shortMessage.set_message((_counter + 1) + " : Skipped");
    else      shortMessage.set_message((_counter + 1) + " : (" + x + ", " + y + ")=" + color2str((int)red(c),(int)green(c),(int)blue(c)));  
    ++_counter;
    if ((max_measure > 0) && (_counter >= max_measure)) {  // 最大計測数に達した場合
      ex_file.flush_data(); //<>//
      if (folder_mode) {
        // フォルダーモードの場合の
        if (next_image() == null) quit_folder_mode (); // もう画像がフォルダー内に無いならば終了
      } else {
        // 単体画像モードの場合はファイル選択ダイアログボックスを表示
        open_fileselector();
      }
    }
    return _counter;
  }
  // 計測取り消し
  int back() {
    if (--_counter < 0) _counter = 0;
    shortMessage.set_message("Back to " + _counter);
    return _counter;
  }
  // 計測のスキップ
  int skip() {
    return measure(-1, -1);
  }
  // 計測済みデータの全クリア
  void clear_alldata() {
    while (measured.length > 0) measured = (PointData[])shorten(measured);
    _counter = 0; // カウンターのリセット
  }
  // 計測済みの全プロットを描画
  void plot_all() {
    strokeWeight(plotMark.lineWidth);
    for (int cnt = 0; cnt < _counter; cnt++) plotMark.plot(measured[cnt].px, measured[cnt].py);
    strokeWeight(1);
  }
  
  // ファイルリストに基づく連続処理
  void open_files(File [] f) {
    files = f;
    f_no  = 0;
    folder_mode = true;

    shortMessage.set_message("[Folder mode]:Start");
    ex_file.flush_data();  // 計測済みデータの書き出し
    if (next_image(true) == null) {
      // 次の画像ファイルが存在しなかった場合
      quit_folder_mode();
    }
  }
  // フォルダーモード
  void folderSelected(File selection) {
    if (selection == null) return;  // キャンセル時

    folder_mode_folder = selection.getAbsolutePath();
    shortMessage.set_message("User select folder = " + ci.folder_mode_folder);
    open_files(selection.listFiles());
  }
  // 次の画像を表示（フォルダーモード）
  PImage next_image() {
    return next_image(false);
  }
  PImage next_image(boolean firsttime) {
    PImage ret = null;
    if (! firsttime) f_no++;  // 初回以外は最初にカウントアップする：次の画像だから
    for (; f_no < files.length; f_no++) {
      if (files[f_no].getPath().endsWith(Export_ext)) continue;  // データファイルは読み飛ばす（ワーニングが少しは減る）
      if (files[f_no].isDirectory())                  continue;  // フォルダーは読み飛ばす
      if ((ret = loadImage(files[f_no].getAbsolutePath())) != null) break;
    }
    if (ret == null) return null;  // 最後のファイルに到達（存在しない）
    ret = load(files[f_no].getAbsolutePath());
    return ret;
  }
  void quit_folder_mode() {
    folder_mode = false;
    img         = null;
    shortMessage.set_message("[Folder mode]:Quit");
  }

  // 画像の拡大率変更
  void rescale(double rate) {
    double pre_scale = scale;
    if (rate == 0.0) {
      if (scale == Initial_scale) return;                 // 初期値で変更なしの場合
      else                        scale = Initial_scale;  // 初期化
    } else {
      scale *= rate;
    }
    // 既に計測済みの場合はスケールを変換する
    if (_counter > 0) {
      int n;
      for (n = 0; n < _counter; n++) {
        measured[n].px = (int)(measured[n].px / pre_scale * scale);
        measured[n].py = (int)(measured[n].py / pre_scale * scale);
      }
    }
    shortMessage.set_message("Change scale " + floorN((float)pre_scale, 4) + " -> " + floorN((float)scale, 4));
    reload();  // 画像の再読み込み
  }
  
  // フィルターの変更
  void change_filter() {
    if (++filter_no >= filter_type.length) filter_no = -1;  //（-1はフィルターなし）
    if (ci.img != null) reload();
  }
  void apply_filter() {
    if (filter_no >= filter_type.length) filter_no = -1;  // 無いはずだが念のため（-1はフィルターなし）

    if (ci.img == null) return;
    if (filter_no >= 0) img.filter(filter_type[filter_no]);
  }
}

// ファイル出力関係
class ExportFile {
  String  dest_filename;
  PrintWriter output;
  
  // ファイル出力時の意図的な遅延
  boolean recording;
  int     delay_rectime;

  void start_recording() {
    recording     = true;
    delay_rectime = millis();
  }
  boolean recording_now() {
    if (millis() - delay_rectime > Delay_rectime) recording = false;  // 計測終了 → 次の動作までの遅延
    return recording;
  }

  ExportFile() {
    dest_filename = "";
    output        = null;
  }
  // ファイルへ出力
  void export_data() {
    int     cnt = 0;
    dest_filename = ci.fullpathname + Export_ext;  // 拡張子の分解は面倒くさいので，単純に拡張子を追加する
    output = createWriter(dest_filename);
    if (output != null) {
      File fp = new File(ci.fullpathname);
      String dateString;
      dateString = year() + nf(month(),2) + nf(day(), 2) + nf(hour(), 2) + nf(minute(), 2) + nf(second(), 2);
      output.println(ci.scale + "\t" + ci.org_width +"\t" + ci.org_height + "\t" + fp.getName() + "\t" + dateString);
      for (cnt = 0; cnt < ci._counter; cnt++) {
        output.print(cnt + "\t" + ci.measured[cnt].px + "\t" + ci.measured[cnt].py);
        output.println("\t" + (int)red(ci.measured[cnt].c) + "\t" + (int)green(ci.measured[cnt].c) + "\t" + (int)blue(ci.measured[cnt].c));
      }
      output.flush();
      output.close();
      shortMessage.set_message("Completed exporting all data to " + dest_filename);
      // 配列の廃棄
      ci.clear_alldata();
      start_recording();  // 意図的な遅延タイマースタート
    } else shortMessage.set_message("Error: Can't output data to file =" + dest_filename);
  }
  // 計測済みデータの出力
  void flush_data() {
    if (ci.img != null && ci._counter > 0) export_data();  // 中断して次のファイルを選択した際に，計測済みのデータを出力する
    ci.img = null;
  }
};

float floorN(float f, int n) {
  return float(floor(f * pow(10, n))) / pow(10, n);
}

// 計測するデータに関するクラス
class PointData {
  int   px, py;  // X, Y座標
  color c;       // 色

  PointData() {
    px = py = -1;    // 特に意味はない
    c       =  0;
  }
  // 計測点の色情報を格納する
  void get_color(PImage img) {
    int idx = px + py * img.width;
    if (idx >= img.pixels.length) return;
    c = img.pixels[idx];
  }
}

// プロットに関するクラス
final String [] plotName = {"none", "cross", "circle", "circlefill", "rect", "rectfill", "scope", "scopefill"}; 

class Plot {
  char mode;     // 0: none, 1: cross,  2: circle, 3: circlefill, 4: rect, 5: rectfill, 6, scope, 7. scopefill
  int  size;
  int  p_color;  // 0, 1, 2
  int  lineWidth;

  Plot() {
    mode = 1;
    size = Plot_size;
    lineWidth = 1;
  }
  // プロットの形状を変更
  void change_shape() {
    if (++mode >= plotName.length) mode = 0;
  }
  String get_shape() {
    return plotName[mode];
  }
  // プロットの色を変更
  void change_color() {
    if (++p_color > Plot_color.length - 1) p_color = 0; 
  }
  // プロットを描く
  // 指定されたプロットを描画
  void plot(int x, int y) {
    stroke(this.red(), this.green(), this.blue());
    noFill();
    switch (mode) {
      case 0 : break;
      case 1 : // cross
        cross(x, y);
        break;
      case 3 : // circlefill
        fill(this.red(), this.green(), this.blue());
      case 2 : // circle 
        circle(x, y);
        break;
      case 5 : // rectfill
        fill(this.red(), this.green(), this.blue());
      case 4 : // rect 
        rectangle(x, y); 
        break;
      case 7 : // scopefill
        fill(this.red(), this.green(), this.blue());
      case 6 : // scope      
        cross(x, y, true);
        circle(x, y);
        break;
      default : mode = 0;
    }
  }
  void cross(int x, int y) {
    cross(x, y, false);
  }
  void cross(int x, int y, boolean large) {
    int s = size;
    if (large) s *= 2;
    line(x - s, y, x + s, y);
    line(x, y - s, x, y + s);
  }
  void circle(int x, int y) {
    ellipse(x, y, size * 2, size * 2);
  }
  void rectangle(int x, int y) {
    rectMode(CENTER);
    rect(x, y, size, size);
    rectMode(CORNER);
  }
  int red() {
    return Plot_color[p_color][0];
  }
  int green() {
    return Plot_color[p_color][1];
  }
  int blue() {
    return Plot_color[p_color][2];
  }
  String color_str() {
    return color2str(this.red(), this.green(), this.blue());
  }
}

String color2str(int r, int g, int b) {
  String s;
  s = "[" + r + "," + g + "," + b + "]";
  return s;
}

// 正の整数をキーボードから入力するためのクラス
class InputInteger {
  boolean inputing_now;
  int     value;
  
  InputInteger() {
    inputing_now = false;
    value        = 0;
  }
  void start() {
    start(0);
  }
  void start(int init) {
    inputing_now = true;
    value        = init;
  }
  void keyin(char c) {
    int code;  // キャラクターコード
    code = int(c) - 48;  // 0 - 9
    if ((code >= 0) && (code <= 9)) {
      add(code);
    } else if (c == BACKSPACE) {
      del();
    } else if (c == ENTER) {
      end();
    } else if (c == ' ') {
      value = 0;
    }
  }
  void end() {
    inputing_now = false;
  }
  void add(int n) {
    value *= 10;
    value += n;
  }
  void del() {
    value /= 10;
  }
}

// ウインドウとコンソールエリアに短いメッセージを一時的に表示するためのクラス
class ShortMessage {
private  int    start;
private  String message;
  ShortMessage() {
    start   = millis();
    message = "";
  }
  void set_message(String str) {
    message = str;
    println(str);
    start   = millis();
  }
  void display_short_message(int x, int y) {
    if (millis() - start < Delay_message) {
      fill(200, 200, 200);
      textSize(16);
      text(message, x, y);
    }
  }
}
